#include <Arduino.h>
#include <Wire.h>
#include <SoftwareSerial.h>
#include "SdsDustSensor.h"
#include <DFRobot_SHT3x.h>
#include <stdarg.h>

// === Définition des broches ===
#define SDS_RX 5
#define SDS_TX 4
#define LORA_RX 13
#define LORA_TX 2
#define SDA_PIN 12
#define SCL_PIN 14

// === Objets capteurs ===
SoftwareSerial SDS_Serial(SDS_RX, SDS_TX);
SoftwareSerial LoRaSerial(LORA_RX, LORA_TX);
SdsDustSensor sds(SDS_Serial);
DFRobot_SHT3x sht3x;

static char recv_buf[512];

// === Fonction AT LoRa-E5 ===
static int at_send_check_response(const char *p_ack, int timeout_ms, const char *p_cmd, ...) {
  va_list args;
  int index = 0;
  int ch;
  unsigned long startMillis;
  memset(recv_buf, 0, sizeof(recv_buf));
  va_start(args, p_cmd);
  char cmd[256];
  vsnprintf(cmd, sizeof(cmd), p_cmd, args);
  va_end(args);
  LoRaSerial.print(cmd);
  delay(200);
  startMillis = millis();
  if (p_ack == NULL) return 0;
  do {
    while (LoRaSerial.available()) {
      ch = LoRaSerial.read();
      if (index < (int)(sizeof(recv_buf) - 1)) {
        recv_buf[index++] = ch;
      }
    }
    if (strstr(recv_buf, p_ack)) return 1;
  } while (millis() - startMillis < (unsigned long)timeout_ms);
  return 0;
}

// === Affichage lisible des mesures ===
void printReadableData(float pm25, float pm10, float tempC, float hum) {
  Serial.println("---- Données Mesurées ----");
  Serial.print("PM2.5 : "); Serial.print(pm25); Serial.println(" µg/m³");
  Serial.print("PM10  : "); Serial.print(pm10); Serial.println(" µg/m³");
  Serial.print("Temp. : "); Serial.print(tempC); Serial.println(" °C");
  Serial.print("Hum.  : "); Serial.print(hum); Serial.println(" %");
  Serial.println("--------------------------");
}

// === Setup (une fois toutes les 5 min) ===
void setup() {
  delay(2000);  // Attendre que le port série soit prêt

  Serial.begin(115200);
  Serial.println("=== Redémarrage ===");

  // Init capteurs
  Wire.begin(SDA_PIN, SCL_PIN);
  sht3x.begin();
  sht3x.startPeriodicMode(sht3x.eMeasureFreq_1Hz);

  SDS_Serial.begin(9600);
  delay(1000);
  sds.begin();
  sds.wakeup();
  delay(10000);


  LoRaSerial.begin(9600);
  delay(200);
  at_send_check_response("+AT: OK", 500, "AT\r\n");
  at_send_check_response("+MODE: TEST", 1000, "AT+MODE=TEST\r\n");

  // Lecture capteurs
  delay(1000);
  PmResult pm = sds.readPm();
  DFRobot_SHT3x::sRHAndTemp_t data = sht3x.readTemperatureAndHumidity();

  // Affichage lisible des mesures
  printReadableData(pm.pm25, pm.pm10, data.TemperatureC, data.Humidity);

  // Payload LoRa
  char hexPayload[32];
  sprintf(hexPayload, "%04X%04X%04X%04X",
          (uint16_t)(pm.pm25 * 100),
          (uint16_t)(pm.pm10 * 100),
          (uint16_t)(data.TemperatureC * 100),
          (uint16_t)(data.Humidity * 100));

  char cmd[64];
  sprintf(cmd, "AT+TEST=TXLRPKT,\"%s\"\r\n", hexPayload);
  at_send_check_response("+TEST: TXLRPKT", 3000, cmd);

  // Mise en veille des capteurs
  sds.sleep();
  sht3x.stopPeriodicMode();
  at_send_check_response("+LOWPOWER: SLEEP", 1000, "AT+LOWPOWER\r\n");

  // Veille profonde pendant 5 minutes
  Serial.println("😴 Veille 5 minutes...");
  Serial.flush();
  ESP.deepSleep(10e6);  // 5 minutes en microsecondes
}

void loop() {
  // vide : jamais utilisé après deepSleep
}
