#include <Arduino.h>
#include <Wire.h>
#include <SoftwareSerial.h>
#include "SdsDustSensor.h"
#include <DFRobot_SHT3x.h>
#include <stdarg.h>

// SDS011 via SoftwareSerial
#define SDS_RX 5   // GPIO5 = D1 (RX de l’ESP ← TX SDS011)
#define SDS_TX 4   // GPIO4 = D2 (TX de l’ESP → RX SDS011)

// LoRa-E5 via SoftwareSerial
#define LORA_RX 13 // GPIO13 = D7 (RX de l’ESP ← TX LoRa)
#define LORA_TX 2  // GPIO2 = D4

// SHT30 via I2C
#define SDA_PIN 12 // GPIO12 = D6
#define SCL_PIN 14 // GPIO14 = D5

// LED intégrée
#define LED_BUILTIN_PIN 2 // GPIO2 = D4

SoftwareSerial SDS_Serial(SDS_RX, SDS_TX);   // SDS011
SoftwareSerial LoRaSerial(LORA_RX, LORA_TX); // LoRa-E5

SdsDustSensor sds(SDS_Serial);
DFRobot_SHT3x sht3x;

static char recv_buf[512];

// Fonction AT (LoRa-E5)
static int at_send_check_response(const char *p_ack, int timeout_ms, const char *p_cmd, ...) {
  va_list args;
  int index = 0;
  int ch;
  unsigned long startMillis;
  memset(recv_buf, 0, sizeof(recv_buf));
  va_start(args, p_cmd);
  char cmd[256];
  vsnprintf(cmd, sizeof(cmd), p_cmd, args);
  va_end(args);
  LoRaSerial.print(cmd);
  Serial.print("-> "); Serial.print(cmd);
  delay(200);
  startMillis = millis();

  if (p_ack == NULL) return 0;

  do {
    while (LoRaSerial.available()) {
      ch = LoRaSerial.read();
      if (index < (int)(sizeof(recv_buf) - 1)) {
        recv_buf[index++] = ch;
      }
      Serial.print((char)ch);
    }
    if (strstr(recv_buf, p_ack)) return 1;
  } while (millis() - startMillis < (unsigned long)timeout_ms);
  return 0;
}

void setup() {
  Serial.begin(115200);

  pinMode(LED_BUILTIN_PIN, OUTPUT);
  digitalWrite(LED_BUILTIN_PIN, HIGH); // Éteint LED intégrée

  // I2C init pour SHT30
  Wire.begin(SDA_PIN, SCL_PIN);
  while (sht3x.begin() != 0) {
    Serial.println("❌ Erreur initialisation SHT30 !");
    delay(1000);
  }
  sht3x.startPeriodicMode(sht3x.eMeasureFreq_1Hz);

  // SDS011 init
  SDS_Serial.begin(9600);
  delay(1000);
  sds.begin();
  sds.wakeup();

  // LoRa-E5 init via AT
  LoRaSerial.begin(9600);
  delay(200);
  at_send_check_response("+AT: OK", 500, "AT\r\n");
  at_send_check_response("+MODE: TEST", 1000, "AT+MODE=TEST\r\n");
}

void printReadableData(float pm25, float pm10, float tempC, float hum) {
  Serial.println("---- Données Mesurées ----");
  Serial.print("PM2.5 : "); Serial.print(pm25); Serial.println(" µg/m³");
  Serial.print("PM10  : "); Serial.print(pm10); Serial.println(" µg/m³");
  Serial.print("Temp. : "); Serial.print(tempC); Serial.println(" °C");
  Serial.print("Hum.  : "); Serial.print(hum); Serial.println(" %");
  Serial.println("--------------------------");
}

void loop() {
  PmResult pm = sds.readPm();
  if (!pm.isOk()) {
    Serial.println("❌ Erreur SDS011 !");
    delay(5000);
    return;
  }

  delay(1100);
  DFRobot_SHT3x::sRHAndTemp_t data = sht3x.readTemperatureAndHumidity();
  if (data.ERR != 0) {
    Serial.println("❌ Erreur lecture SHT30 !");
    delay(5000);
    return;
  }

  printReadableData(pm.pm25, pm.pm10, data.TemperatureC, data.Humidity);

  // Payload hexadécimal
  char hexPayload[32];
  sprintf(hexPayload, "%04X%04X%04X%04X",
          (uint16_t)(pm.pm25 * 100),
          (uint16_t)(pm.pm10 * 100),
          (uint16_t)(data.TemperatureC * 100),
          (uint16_t)(data.Humidity * 100));

  Serial.print("Payload hex envoyé : ");
  Serial.println(hexPayload);

  char cmd[64];
  sprintf(cmd, "AT+TEST=TXLRPKT,\"%s\"\r\n", hexPayload);
  int ok = at_send_check_response("+TEST: TXLRPKT", 3000, cmd);

  if (ok) {
    Serial.println("✅ Données envoyées avec succès !");
    // LED clignote une fois rapidement
    digitalWrite(LED_BUILTIN_PIN, LOW); // Allume
    delay(200);
    digitalWrite(LED_BUILTIN_PIN, HIGH); // Éteint
  } else {
    Serial.println("❌ Échec de l'envoi.");
    // LED clignote 3 fois lentement
    for (int i = 0; i < 3; i++) {
      digitalWrite(LED_BUILTIN_PIN, LOW);
      delay(300);
      digitalWrite(LED_BUILTIN_PIN, HIGH);
      delay(300);
    }
  }

  delay(60000); // Attente avant prochaine mesure
}
