#include <Arduino.h>
#include <Wire.h>
#include <SoftwareSerial.h>
#include "SdsDustSensor.h"
#include <DFRobot_SHT3x.h>
#include <stdarg.h>

// SDS011 via SoftwareSerial
#define SDS_RX 5
#define SDS_TX 4

// LoRa-E5 via SoftwareSerial
#define LORA_RX 13
#define LORA_TX 2

// SHT30 via I2C
#define SDA_PIN 12
#define SCL_PIN 14

SoftwareSerial SDS_Serial(SDS_RX, SDS_TX);
SoftwareSerial LoRaSerial(LORA_RX, LORA_TX);

SdsDustSensor sds(SDS_Serial);
DFRobot_SHT3x sht3x;

static char recv_buf[512];

// TTN identifiers
const char* DevEUI = "70B3D57ED00630C5";
const char* AppEUI = "000000000000000";
const char* AppKey = "FFC756C21917F54E3993167F8E68AB22";

// Fonction AT améliorée
static int at_send_check_response(const char *p_ack, int timeout_ms, const char *p_cmd, ...) {
  va_list args;
  int index = 0;
  int ch;
  unsigned long startMillis;
  memset(recv_buf, 0, sizeof(recv_buf));
  va_start(args, p_cmd);
  char cmd[256];
  vsnprintf(cmd, sizeof(cmd), p_cmd, args);
  va_end(args);
  LoRaSerial.print(cmd);
  Serial.print("-> "); Serial.print(cmd);
  delay(200);
  startMillis = millis();

  do {
    while (LoRaSerial.available()) {
      ch = LoRaSerial.read();
      if (index < (int)(sizeof(recv_buf) - 1)) {
        recv_buf[index++] = ch;
      }
      Serial.print((char)ch);
    }

    // On accepte plusieurs types de réponses positives
    if (strstr(recv_buf, "+CMSGHEX: Done") || strstr(recv_buf, "+MSG: Done") || (p_ack && strstr(recv_buf, p_ack)))
      return 1;

  } while (millis() - startMillis < (unsigned long)timeout_ms);

  Serial.println("Réponse complète LoRaE5 (échec) :");
  Serial.println(recv_buf);
  return 0;
}

void printReadableData(float pm25, float pm10, float tempC, float hum) {
  Serial.println("---- Données Mesurées ----");
  Serial.print("PM2.5 : "); Serial.print(pm25); Serial.println(" µg/m³");
  Serial.print("PM10  : "); Serial.print(pm10); Serial.println(" µg/m³");
  Serial.print("Temp. : "); Serial.print(tempC); Serial.println(" °C");
  Serial.print("Hum.  : "); Serial.print(hum); Serial.println(" %");
  Serial.println("--------------------------");
}

void setup() {
  Serial.begin(115200);

  Wire.begin(SDA_PIN, SCL_PIN);
  while (sht3x.begin() != 0) {
    Serial.println("❌ Erreur initialisation SHT30 !");
    delay(1000);
  }
  sht3x.startPeriodicMode(sht3x.eMeasureFreq_1Hz);

  SDS_Serial.begin(9600);
  delay(1000);
  sds.begin();
  sds.wakeup();

  LoRaSerial.begin(9600);
  delay(200);

  at_send_check_response("+AT: OK", 500, "AT\r\n");
  at_send_check_response("+MODE: LWOTAA", 1000, "AT+MODE=LWOTAA\r\n");

  at_send_check_response("+ID: DevEui", 1000, "AT+ID=DevEui,\"%s\"\r\n", DevEUI);
  at_send_check_response("+ID: AppEui", 1000, "AT+ID=AppEui,\"%s\"\r\n", AppEUI);
  at_send_check_response("+KEY: APPKEY", 1000, "AT+KEY=APPKEY,\"%s\"\r\n", AppKey);

  at_send_check_response("+DR: EU868", 1000, "AT+DR=EU868\r\n");

  Serial.println("🔄 Tentative de join...");

  while (!at_send_check_response("+JOIN: Network joined", 10000, "AT+JOIN\r\n")) {
    Serial.println("❌ Join échoué ! Nouvelle tentative dans 5 secondes...");
    delay(5000);
  }

  Serial.println("✅ Rejoint avec succès !");
}

void loop() {
  PmResult pm = sds.readPm();
  if (!pm.isOk()) {
    Serial.println("❌ Erreur SDS011 !");
    delay(5000);
    return;
  }

  delay(1100);
  DFRobot_SHT3x::sRHAndTemp_t data = sht3x.readTemperatureAndHumidity();
  if (data.ERR != 0) {
    Serial.println("❌ Erreur lecture SHT30 !");
    delay(5000);
    return;
  }

  printReadableData(pm.pm25, pm.pm10, data.TemperatureC, data.Humidity);

  char hexPayload[32];
  sprintf(hexPayload, "%04X%04X%04X%04X",
          (uint16_t)(pm.pm25 * 100),
          (uint16_t)(pm.pm10 * 100),
          (uint16_t)(data.TemperatureC * 100),
          (uint16_t)(data.Humidity * 100));

  Serial.print("Payload hex envoyé : ");
  Serial.println(hexPayload);

  char cmd[64];
  sprintf(cmd, "AT+CMSGHEX=\"%s\"\r\n", hexPayload);

  int ok = at_send_check_response("+CMSGHEX: Done", 10000, cmd);

  if (ok)
    Serial.println("✅ Données envoyées avec succès !");
  else
    Serial.println("❌ Échec de l'envoi.");

  delay(60000);
}
